{ sharing sourcecode to Pascal Game Development members by Micronix / TRSI }

PROGRAM OpenGLApp;

uses
  Windows,
  Messages,
  dglOpenGL,
  bmp,
  ms3d;

CONST
  Wnd_Title        = 'OpenGL Sample';
  Error_Message_00 = 'Release of DC and RC failed';
  Error_Message_01 = 'Release of rendering context failed';
  Error_Message_02 = 'Release of device context failed';
  Error_Message_03 = 'Unable to destroy window';
  Error_Message_04 = 'Unable to unregister window class';
  Error_Message_05 = 'Failed to register the window class';
  Error_Message_06 = 'Unable to activate fullscreen';
  Error_Message_07 = 'Unable to create window';
  Error_Message_08 = 'Unable to get a device context';
  Error_Message_09 = 'Unable to find a suitable pixel format';
  Error_Message_10 = 'Unable to set the pixel format';
  Error_Message_11 = 'Unable to create rendering context';
  Error_Message_12 = 'Unable to activate rendering context';
  { screen settings }
  XRes             = 640;
  YRes             = 480;
  PixelDepth       = 32;
  { timer to calculate frames per second }
  FPS_Timer        = 1;
  { calculate FPS every 1000 ms }
  FPS_Interval     = 1000;
VAR
  DemoStart,
  LastTime   : DWORD;
  { Counter for FPS }
  FPSCount          : INTEGER = 0;
  { mouse button down }
  MouseButton       : INTEGER = -1;
  { Is the mouse moving? }
  MouseMove         : BOOLEAN = FALSE;
  FullScreen        : BOOLEAN = FALSE;
  OpenGLInitialized : BOOLEAN = FALSE;
  { global window handle }
  h_Wnd             : HWND;
  { global device context }
  h_Dc              : HDC;
  { OpenGL rendering context }
  h_Rc              : HGLRC;
  { Holds keystrokes }
  Keys              : ARRAY [ 0 .. 255 ] OF BOOLEAN;
  { elapsed time between frames }
  ElapsedTime,
  { Current X and Y position where the mouse was clicked }
  XCoord,
  YCoord            : INTEGER;
  { Current X and Y position of the cursor }
  MoveXCoord,
  MoveYCoord        : GLfloat;
  LightAmbient : array[0..3] of Single = (0.5, 0.5, 0.5, 1);
  LightDiffuse : array[0..3] of Single = (1.0, 0.6, 0.5, 1);
  LightSpecular: array[0..3] of Single = (0.0, 0.0, 1.0, 1);
  LightDirection:array[0..2] of Single = (-1, 0, 0);
  LightPosition: array[0..3] of Single = (50.0, 50.0, 50.0, 1);
  gray         : array[0..3] of Single = (1.0, 1.0, 1.0, 1);
  model : TAnimatedModelGL;
  ControlPoints: array[0..19, 0..5] of Single = (
    (000.00, 000.00, -04.50,   9.70,  -50.50, 000.00 ),
    (000.00, 000.00, -13.51,  67.70,    9.51, 000.00 ),
    (000.00, 000.00, -23.06,  78.16,  107.06, 000.00 ),
    (000.00, 000.00, -32.93,  63.90, -156.93, 000.00 ),
    (000.00, 000.00, -43.49,  46.63,  -71.49, 000.00 ),
    (000.00, 000.00, -53.14,   9.88,   -2.14, 000.00 ),
    (000.00, 000.00, -63.75,  42.82,   73.75, 000.00 ),
    (000.00, 000.00, -23.30,  24.24,  153.30, 000.00 ),
    (000.00, 000.00, -13.55,   1.10, -130.55, 000.00 ),
    (000.00, 000.00, -22.59,  60.50,  -69.59, 000.00 ),
    (000.00, 000.00, -33.84,  63.80,   56.84, 000.00 ),
    (000.00, 000.00, -43.60,  94.25,  170.60, 000.00 ),
    (000.00, 000.00, -03.19,  59.02,   93.19, 000.00 ),
    (000.00, 000.00, -13.53,  44.36,  162.53, 000.00 ),
    (000.00, 000.00, -04.82,  27.18,  -74.82, 000.00 ),
    (000.00, 000.00, -23.28,  92.87,  -23.28, 000.00 ),
    (000.00, 000.00, -03.78,  65.47,   38.78, 000.00 ),
    (000.00, 000.00, -12.47,  69.01,   89.47, 000.00 ),
    (000.00, 000.00, -23.34,   6.39, -170.34, 000.00 ),
    (000.00, 000.00, -13.71,  23.20, -104.71, 000.00 ) );
  Tex1: glUint;
  am: SINGLE = 0;

{$R *.RES}

function Interpolate(P1, P2, Amt: Single): Single;
begin
  Amt := (1 - cos (Amt * Pi)) / 2;
  Result := P1 * (1 - Amt) + P2 * Amt;
end;

{---------------------------------------------------------------------------}
{  initialise OpenGL scene                                                  }
{---------------------------------------------------------------------------}
FUNCTION glInit : BOOLEAN;
BEGIN
  glEnable(GL_TEXTURE_2D);	       // Aktiviert Texture Mapping
  glShadeModel(GL_SMOOTH);	       // Aktiviert weiches Shading
  glClearColor(0.0, 0.0, 0.0, 0.5);    // Bildschirm lschen (schwarz)
  glClearDepth(1.0);		       // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);	       // Aktiviert Depth Testing
  glDepthFunc(GL_LEQUAL);	       // Bestimmt den Typ des Depth Testing
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);  // Qualitativ bessere Koordinaten Interpolation
  glEnable(GL_TEXTURE_2D);
  { load the textures }
  LoadTexture( 'Floor.bmp', Tex1  );
  glLightfv(GL_LIGHT1, GL_AMBIENT, @LightAmbient);
  glLightfv(GL_Light1, GL_DIFFUSE, @LightDiffuse);
  glLightfv(GL_LIGHT1, GL_POSITION, @LightPosition);
  glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, @LightDirection);
  glLightfv(GL_LIGHT1, GL_SPECULAR, @LightSpecular);
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHTING);
  glEnable(GL_NORMALIZE);
  model := ms3dLoadAnimatedModel('models\ninja.ms3d');
  model.mintime:=40*206;        { start with idle position }
  model.time:=   40*206;
  model.MaxTime:=40*250;
  LastTime := ElapsedTime;
  model.ShowBones := FALSE;
  model.ShowBoundingBox :=FALSE;
  Result := TRUE
END;
{---------------------------------------------------------------------------}
{  draw OpenGL scene                                                        }
{---------------------------------------------------------------------------}
FUNCTION glDraw : BOOLEAN;
var t : integer;
VAR co2: INTEGER;
 var
  P1, P2: Integer;
  rx,ry,rz,tx,ty,tz: SINGLE;
begin
  glClearColor(0,0,0,0);
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity;
  P1 := Trunc(Am) mod Length(ControlPoints);
  P2 := (P1 + 1) mod Length(ControlPoints);
  tx := Interpolate(ControlPoints[P1, 0], ControlPoints[P2, 0], Frac(Am));
  ty := Interpolate(ControlPoints[P1, 1], ControlPoints[P2, 1], Frac(Am));
  tz := Interpolate(ControlPoints[P1, 2], ControlPoints[P2, 2], Frac(Am));
  Rx := Interpolate(ControlPoints[P1, 3], ControlPoints[P2, 3], Frac(Am));
  RY := Interpolate(ControlPoints[P1, 4], ControlPoints[P2, 4], Frac(Am));
  RZ := Interpolate(ControlPoints[P1, 5], ControlPoints[P2, 5], Frac(Am));
  glBindTexture    ( GL_TEXTURE_2D, Tex1 );
  glTranslatef(tx, ty, tz);
  glRotatef(Rx, 1.0, 0.0, 0.0);
  glRotatef(Ry, 0.0, 1.0, 0.0);
  glRotatef(Rz, 0.0, 0.0, 1.0);
  am:=am+0.005;
  glRotatef(-90,1,0,0);
  glBegin(GL_QUADS);
    glTexCoord2f(00,00);glVertex2f(-30,-30);
    glTexCoord2f(10,00);glVertex2f( 30,-30);
    glTexCoord2f(10,10);glVertex2f( 30, 30);
    glTexCoord2f(00,10);glVertex2f(-30, 30);
  glEnd;
  glenable(GL_TEXTURE_2D);
  glRotatef ( 90,1,0,0 );
  glRotatef ( 45,0,1,0 );
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, @gray);
  t := ElapsedTime - LastTime;
  model.AdvanceAnimation( t div 2 );
  model.render;
  glFlush();
  Result := TRUE
END;

{---------------------------------------------------------------------------}
{  handle window resize                                                     }
{---------------------------------------------------------------------------}
PROCEDURE glResizeWnd ( Width, Height : GLsizei );
BEGIN
  { prevent divide by zero exception }
  IF Height = 0
    THEN
      Height := 1;
  { set the viewport for the OpenGL window }
  glViewport ( 0, 0, Width, Height );
  { change matrix mode to projection }
  glMatrixMode ( GL_PROJECTION );
  { reset view }
  glLoadIdentity;
  { do the perspective calculations. last value = max clipping depth }
  gluPerspective ( 45.0, Width / Height, 1.0, 100.0 );
  { return to the modelview matrix }
  glMatrixMode ( GL_MODELVIEW );
  { reset view }
  glLoadIdentity;
END;

{---------------------------------------------------------------------------}
{  processes all the keystrokes                                             }
{---------------------------------------------------------------------------}
PROCEDURE ProcessKeys;
BEGIN
   { Punch and swipe sword }
   if (keys[$31]) then BEGIN model.mintime:=40*32;  model.MaxTime:=40*44; model.restart END;
   { Swipe and spin sword }
   if (keys[$32]) then BEGIN model.mintime:=40*45;  model.MaxTime:=40*59; model.restart END;
   { Overhead twohanded downswipe }
   if (keys[$33]) then BEGIN model.mintime:=40*60;  model.MaxTime:=40*68; model.restart END;
   { Forward kick }
   if (keys[$34]) then BEGIN model.mintime:=40*73;  model.MaxTime:=40*83; model.restart END;
   { Side Kick }
   if (keys[$35]) then BEGIN model.mintime:=40*126;  model.MaxTime:=40*133; model.restart END;
   { Jump }
   if (keys[$36]) then BEGIN model.mintime:=40*94;  model.MaxTime:=40*102; model.restart END;
   { Death 1 - Fall back onto ground }
   if (keys[$37]) then BEGIN model.mintime:=40*166;  model.MaxTime:=40*173; model.restart END;
   { display bones }
   if (keys[$38]) THEN BEGIN model.ShowBones:=NOT model.ShowBones; keys[$38]:=FALSE END;
   { display bones }
   if (keys[$39]) then BEGIN model.ShowBoundingBox:=NOT model.ShowBoundingBox; keys[$39]:=FALSE END;
END;

{---------------------------------------------------------------------------}
{  processes all the mouse move                                             }
{---------------------------------------------------------------------------}
PROCEDURE ProcessMouseMove;
BEGIN
  IF MouseMove
    THEN
      BEGIN
        { add code here to execute on mouse move }
        MouseMove := False
      END
END;

{---------------------------------------------------------------------------}
{  processes all the mouse clicks                                           }
{---------------------------------------------------------------------------}
PROCEDURE ProcessMouseClick;
BEGIN
  CASE MouseButton OF
    { left mouse button }
    1 : BEGIN
          { add code to execute on mouse click }
          MouseButton := 0
          { to use this procedure as a mouse down event remove line above }
        END;
    { right mouse button }
    2 : BEGIN
          { add code to execute on mouse click }
          MouseButton := 0
          { to use this procedure as a mouse down event remove line above }
        END;
    { middle mouse button }
    3 : BEGIN
          { add code to execute on mouse click }
          MouseButton := 0
          { to use this procedure as a mouse down event remove line above }
        END
  END
END;

{---------------------------------------------------------------------------}
{  convert integer to string                                                }
{---------------------------------------------------------------------------}
FUNCTION IntToStr ( Num : INTEGER ) : STRING;
BEGIN
  Str ( Num, Result )
END;

{---------------------------------------------------------------------------}
{  determines the application's response to the messages received           }
{---------------------------------------------------------------------------}
FUNCTION WndProc ( hWnd: HWND; Message: UINT;
                   wParam: WPARAM; lParam: LPARAM): LResult; stdcall;
BEGIN
  IF Message = WM_SYSCOMMAND
    THEN
      CASE wParam OF
        SC_SCREENSAVE, SC_MONITORPOWER:
          BEGIN
            Result := 0;
            Exit
          END
      END;
  CASE Message OF
    WM_ACTIVATE :
      BEGIN
        IF HiWord ( wParam ) = 0
          THEN
            OpenGLInitialized := TRUE;
        Result := 0
      END;
    WM_CLOSE :
      BEGIN
        PostQuitMessage ( 0 );
        Result := 0
      END;
    { set the pressed key to TRUE so we can check if its pressed }
    WM_KEYDOWN :
      BEGIN
        Keys [ wParam ] := TRUE;
        Result := 0
      END;
    { set the pressed key to FALSE so we can check if its pressed }
    WM_KEYUP :
      BEGIN
        Keys[ wParam ] := FALSE;
        Result := 0
      END;
    WM_MOUSEMOVE:
      BEGIN
        MouseMove := TRUE;
        MoveXCoord := LOWORD ( lParam );
        MoveYCoord := HIWORD ( lParam );
        Result := 0
      END;
    WM_LBUTTONDOWN:
      BEGIN
        ReleaseCapture;
        SetCapture ( h_Wnd );
        MouseButton := 1;
        XCoord := LOWORD ( lParam );
        YCoord := HIWORD ( lParam );
        Result := 0
      END;
    WM_RBUTTONDOWN:
      BEGIN
        ReleaseCapture;
        SetCapture ( h_Wnd );
        MouseButton := 2;
        XCoord := LOWORD ( lParam );
        YCoord := HIWORD ( lParam );
        Result := 0
      END;
    WM_MBUTTONDOWN:
      BEGIN
        ReleaseCapture;
        SetCapture ( h_Wnd );
        MouseButton := 3;
        XCoord := LOWORD ( lParam );
        YCoord := HIWORD ( lParam );
        Result := 0
      END;
    WM_LBUTTONUP,
    WM_RBUTTONUP,
    WM_MBUTTONUP:
      BEGIN
        ReleaseCapture;
        MouseButton := 0;
        XCoord := 0;
        YCoord := 0;
        Result := 0
      END;
    WM_SIZE :
      BEGIN
        IF OpenGLInitialized
          THEN
            glResizeWnd( LOWORD ( lParam ), HIWORD ( lParam ) );
        Result := 0
      END;
    WM_TIMER :
      IF wParam = FPS_TIMER
        THEN
          BEGIN
            FPSCount := Round( FPSCount * 1000 / FPS_Interval );
            SetWindowText ( h_Wnd, PChar ( WND_TITLE + '   [' +
              IntToStr ( FPSCount ) + ' FPS]' ) );
            FPSCount := 0;
            Result := 0
          END;
    ELSE
      Result := DefWindowProc( hWnd, Message, wParam, lParam )
    END
END;

{---------------------------------------------------------------------------}
{  display error-message                                                    }
{---------------------------------------------------------------------------}
PROCEDURE ThrowError ( ThrowString : PCHAR );
BEGIN
  MessageBox (0, ThrowString, 'ERROR', MB_OK OR MB_ICONERROR )
END;

{---------------------------------------------------------------------------}
{  properly destroys the window created at startup (no memory leaks)        }
{---------------------------------------------------------------------------}
PROCEDURE glKillWnd;
CONST
  pnil : POINTER = NIL;
BEGIN
  IF FullScreen
    THEN
      { Change back to non fullscreen }
      ChangeDisplaySettings ( devmode ( pnil^ ), 0 );
  { deactivate rendering context }
  IF NOT wglMakeCurrent( h_Dc, 0 )
    THEN
      ThrowError ( Error_Message_00 );
  { attempts to delete the rendering context }
  IF NOT wglDeleteContext( h_Rc )
    THEN
      BEGIN
        ThrowError ( Error_Message_01 );
        h_Rc := 0
      END;
  { attemps to release the device context }
  IF ( h_Dc > 0 ) AND ( releaseDC ( h_Wnd, h_Dc ) = 0 )
    THEN
      BEGIN
        ThrowError ( Error_Message_02 );
        h_Dc := 0
      END;
  { attempts to destroy the window }
  IF ( h_Wnd <> 0 ) AND ( NOT destroywindow ( h_Wnd ) )
    THEN
      BEGIN
        ThrowError ( Error_Message_03 );
        h_Wnd := 0
      END;
  { attempts to unregister the window class }
  IF ( NOT UnregisterClass ( 'OpenGL', hInstance ) )
    THEN
      ThrowError ( Error_Message_04 )
END;

{---------------------------------------------------------------------------}
{  creates the window and attaches a OpenGL rendering context to it         }
{---------------------------------------------------------------------------}
FUNCTION glCreateWnd : BOOLEAN stdcall;
VAR
  wndClass         : TWndclass;                { window class }
  dwStyle,                                     { window styles }
  dwExStyle        : DWORD;                    { extended window styles }
  dmScreenSettings : Devmode;                  { screen settings }
  Pixelformat      : TGLuint;
  h_Instance       : HINST;                    { Current instance }
  pfd              : pixelformatdescriptor;    { GL window settings }
BEGIN
  { get an instance for our window }
  h_instance        := GetModuleHandle ( NIL );
  { set up the window class }
  ZeroMemory ( @wndClass, SizeOf ( wndClass ) );
  WITH wndClass DO
    BEGIN
      style         := CS_HREDRAW OR     { redraws window if length changes }
                       CS_VREDRAW OR     { redraws window if height changes }
                       CS_OWNDC;         { Unique device context for window }
      { set the window procedure to our func WndProc }
      lpfnWndProc   := @WndProc;
      hInstance     := h_Instance;
      hCursor       := LoadCursor ( 0, IDC_ARROW );
      lpszClassName := 'OpenGL'
    END;
  { attemps to register the window class }
  IF  RegisterClass ( wndClass ) = 0
    THEN
      BEGIN
        ThrowError ( Error_Message_05 );
        Result := FALSE;
        Exit
      END;
  { Change to fullscreen if so desired }
  IF FullScreen
    THEN
      BEGIN
        ZeroMemory ( @dmScreenSettings, SizeOf ( dmScreenSettings ) );
        WITH dmScreensettings DO
          BEGIN
            dmSize       := SizeOf ( dmScreenSettings );
            dmPelsWidth  := XRes;                            { window width }
            dmPelsHeight := YRes;                           { window height }
            dmBitsPerPel := PixelDepth;                { Window color depth }
            dmFields     := DM_BITSPERPEL OR DM_PELSWIDTH OR DM_PELSHEIGHT
        END;
        { attemps to change screen mode to fullscreen }
        IF ( ChangeDisplaySettings ( dmScreenSettings, CDS_FULLSCREEN ) ) =
             DISP_CHANGE_FAILED
          THEN
            BEGIN
              ThrowError ( Error_Message_06 );
              FullScreen := FALSE;
            END
      END;
  IF FullScreen
    THEN
      BEGIN
        dwStyle   := WS_POPUP OR                   { creates a popup window }
                     WS_CLIPSIBLINGS OR { don't draw within sibling windows }
                     WS_CLIPCHILDREN;     { don't draw within child windows }
        dwExStyle := WS_EX_APPWINDOW;
        { disable mouse-cursor }
        ShowCursor ( FALSE )
      END
    ELSE
      BEGIN
        dwStyle   := WS_OVERLAPPEDWINDOW OR    { creates overlapping window }
                     WS_CLIPSIBLINGS OR { don't draw within sibling windows }
                     WS_CLIPCHILDREN;     { don't draw within child windows }
        dwExStyle := WS_EX_APPWINDOW OR                  { top level window }
                     WS_EX_WINDOWEDGE           { Border with a raised edge }
      END;
  { attempt to create the actual window }
  H_wnd := CreateWindowEx ( dwExStyle, 'OpenGL', Wnd_Title, dwStyle, 0, 0,
                            XRes, YRes, 0, 0, hinstance, NIL );
  IF h_Wnd = 0
    THEN
      BEGIN
        { undo all the settings we've changed }
        glKillWnd;
        ThrowError ( Error_Message_07 );
        Result:=FALSE;
        Exit
      END;
  { Settings of Pixel Format Descriptor for the OpenGL window }
  WITH pfd DO
    BEGIN
      nSize           := SizeOf ( PIXELFORMATDESCRIPTOR );    { Size Of PDF }
      nVersion        := 1;                { version of this data structure }
      dwFlags         := PFD_DRAW_TO_WINDOW OR          { drawing to window }
                         PFD_SUPPORT_OPENGL OR             { OpenGL drawing }
                         PFD_DOUBLEBUFFER;               { double buffering }
      iPixelType      := PFD_TYPE_RGBA;                 { RGBA color format }
      cColorBits      := PixelDepth;                   { OpenGL color depth }
      cRedBits        := 0;                        { number of red bitplane }
      cRedShift       := 0;                 { shift count for red bitplanes }
      cGreenBits      := 0;                     { number of green bitplanes }
      cGreenShift     := 0;               { shift count for green bitplanes }
      cBlueBits       := 0;                      { number of blue bitplanes }
      cBlueShift      := 0;                { shift count for blue bitplanes }
      cAlphaBits      := 0;                                 { not supported }
      cAlphaShift     := 0;                                 { not supported }
      cAccumBits      := 0;                        { no accumulation buffer }
      cAccumRedBits   := 0;                { number of red bits in a-buffer }
      cAccumGreenBits := 0;              { number of green bits in a-buffer }
      cAccumBlueBits  := 0;               { number of blue bits in a-buffer }
      cAccumAlphaBits := 0;              { number of alpha bits in a-buffer }
      cDepthBits      := 16;                    { depth of the depth buffer }
      cStencilBits    := 0;                    { turn on/off stencil buffer }
      cAuxBuffers     := 0;                                 { not supported }
      iLayerType      := PFD_MAIN_PLANE;                          { ignored }
      bReserved       := 0;         { number of overlay and underlay planes }
      dwLayerMask     := 0;                                       { ignored }
      dwVisibleMask   := 0;           { transparent color of underlay plane }
      dwDamageMask    := 0                                        { ignored }
    END;
  h_Dc := GetDC ( h_Wnd );
  IF h_Dc = 0
    THEN
      BEGIN
        glKillWnd;
        ThrowError ( Error_Message_08 );
        Result := FALSE;
        Exit
      END;
  { attempts to find the pixel format supported by a device context that is }
  { the best match to a given pixel format specification }
  PixelFormat := ChoosePixelFormat ( h_Dc, @pfd );
  IF PixelFormat = 0
    THEN
      BEGIN
        glKillWnd;
        ThrowError ( Error_Message_09 );
        Result := FALSE;
        Exit
      END;
  { sets the specified device context's pixel format to the format }
  { specified by the pixel format }
  IF NOT SetPixelFormat( h_Dc,PixelFormat,@pfd )
    THEN
      BEGIN
        glKillWnd;
        ThrowError ( Error_Message_10 );
        Result := FALSE;
        Exit
      END;
  { create a OpenGL rendering context }
  h_Rc := wglCreateContext ( h_Dc );
  IF h_Rc = 0
    THEN
      BEGIN
      glKillWnd;
      ThrowError ( Error_Message_11 );
      Result := FALSE;
      Exit
    END;
  { activate rendering context }
  IF NOT wglMakeCurrent ( h_Dc, h_Rc )
    THEN
      BEGIN
        glKillWnd;
        ThrowError ( Error_Message_11 );
        Result := FALSE;
        Exit
      END;
  { read and assign extentions }
  ReadExtensions;
  ReadImplementationProperties;
  { initializes the timer used to calculate the FPS }
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);
  { settings to ensure that the window is the topmost window }
  ShowWindow          ( h_Wnd, SW_SHOW );
  SetForegroundWindow ( h_Wnd );
  SetFocus            ( h_Wnd );
  { ensure the OpenGL window is resized properly }
  glResizeWnd         ( XRes, YRes );
  IF NOT glInit
    THEN
      BEGIN
        glKillWnd;
        ThrowError ( Error_Message_12 );
        Result := FALSE;
        Exit
      END;
  Result := TRUE
END;

{---------------------------------------------------------------------------}
{  Main message loop for the application                                    }
{---------------------------------------------------------------------------}
FUNCTION WinMain ( hInstance, hPrevInstance: HINST;
		   lpCmdLine: PChar; nCmdShow: INTEGER): INTEGER; stdcall;
VAR
  Finished   : BOOLEAN;
  msg        : TMsg;
BEGIN
  finished:=false;
  IF MessageBox ( 0, 'Would You Like To Run In FullScreen Mode?',
    'Start FullScreen', MB_YESNO OR MB_ICONQUESTION ) = IDYES
    THEN
      FullScreen := TRUE;
  { dglOpenGl Initialization }
  InitOpenGL;
  { perform application initialization }
  IF NOT glCreateWnd
    THEN
      BEGIN
        Result := 0;
        Exit
      END;
  { get time when demo started }
  DemoStart := GetTickCount;
  { Main message loop }
  WHILE NOT Finished DO
    { Check if there is a message for this window }
    IF PeekMessage ( msg, 0, 0, 0, PM_REMOVE )
      THEN
        BEGIN
          { If WM_QUIT message received }
          IF msg.Message = WM_QUIT
            { then we are Finished }
            THEN
              Finished := TRUE
            { Else translate and dispatch the message to this window }
            ELSE
              BEGIN
                TranslateMessage ( msg );
                DispatchMessage ( msg )
              END
        END
      ELSE
        BEGIN
          { increment FPS counter }
          Inc ( FPSCount );
          LastTime := ElapsedTime;
          { Calculate Elapsed Time }
          ElapsedTime := GetTickCount - DemoStart;
          { average it out for smoother movement }
          ElapsedTime := ( LastTime + ElapsedTime ) DIV 2;
          { If user pressed ESC then set finised TRUE }
          IF OpenGLInitialized AND NOT glDraw OR Keys [ VK_ESCAPE ]
            THEN
              Finished := TRUE
            ELSE
              SwapBuffers ( h_Dc );
          IF Keys [ VK_F1 ]
            THEN
              { toggle fullscreen }
              BEGIN
                Keys [ VK_F1 ] := FALSE;
                glKillWnd;
                FullScreen := NOT FullScreen;
                IF NOT glCreateWnd
                  THEN
                    Result := 0
              END;
          { check for any other key pressed }
          ProcessKeys;
          { check for mouse move }
          ProcessMouseMove;
          { check for mouse click }
          ProcessMouseClick
        END;
  glKillWnd;
  Result := msg.wParam
END;

BEGIN
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow )
END.