// 15.09.2012
// after implementation of bones- and joints visualisation
// and adding a bounding box around the animated character
// for simple and fast collision detection i removed comments
// and gave the source code again a proper structure

UNIT ms3d;

INTERFACE

USES
  dglOpenGL;

TYPE
  ThreeSingles           = ARRAY [ 0..2 ] OF SINGLE;
  FourSingles            = ARRAY [ 0..3 ] OF SINGLE;
  ThreeWords             = ARRAY [ 0..2 ] OF WORD;
  ThreeThreeSingles      = ARRAY [ 0..2 ] OF ThreeSingles;
  TMatrix                = ARRAY [ 0..3 ] OF ARRAY [0..3] OF SINGLE;
  TModelGLMatrix         = TMatrix;
  TModelVertex           = RECORD
                             Position              : ThreeSingles;
                             BoneID                : BYTE
                           END;
  TModelTriangle         = RECORD
                             VertexIndices         : ThreeWords;
                             VertexNormals         : ThreeThreeSingles;
                             s, t                  : ThreeSingles
                           END;
  TModelMaterial         = RECORD
                             Ambient,
                             Diffuse,
                             Specular,
                             Emissive              : FourSingles;
                             Shininess,
                             Transparency          : SINGLE;
                             Texture               : gluint
                           END;
  TModelGroup            = RECORD
                             MaterialIndex         : BYTE;
                             nTriangles            : WORD;
                             TriangleIndices       : ARRAY OF WORD
                           END;
  TModelKeyframe         = RECORD
                             JointIndex            : WORD;
                             Time                  : SINGLE;                    //in ms
                             Parameter             : ThreeSingles
                           END;
  TModelJoint            = RECORD
                             LocalRotation,
                             LocalTranslation      : ThreeSingles;
                             AbsoluteMatrix,
                             RelativeMatrix,
                             FinalMatrix           : TModelGLMatrix;
                             CurTransKeyframe,
                             CurRotKeyframe        : WORD;
                             Parent                : INTEGER;
                             nRotationKeyframes,
                             nTranslationKeyframes : WORD;
                             RotationKeyframes,
                             TranslationKeyframes  : ARRAY OF TModelKeyframe
                           END;
  TModelGLRenderResult   = ( RENDERED_OKAY, MATERIAL_NOT_FOUND );
  TStaticModelGL         = CLASS
                             PUBLIC
                             numGroups,
                             numMaterials,
                             numTriangles,
                             numVertices         : WORD;
                             Vertices            : ARRAY OF TModelVertex;
                             Triangles           : ARRAY OF TModelTriangle;
                             Groups              : ARRAY OF TModelGroup;
                             Materials           : ARRAY OF TModelMaterial;
                             FUNCTION Render : TModelGLRenderResult;
                           END;
  TAnimatedModelGL       = CLASS ( TStaticModelGL )
                             PUBLIC
                             Time,
                             MinTime,
                             MaxTime             : INTEGER;
                             BoundingBoxMin,
                             BoundingBoxMax      : ThreeSingles;
                             ShowBones,
                             ShowBoundingBox     : BOOLEAN;
                             numJoints           : WORD;
                             Joints              : ARRAY OF TModelJoint;
                             FUNCTION Render : TModelGLRenderResult;
                             PROCEDURE DisplayBones;
                             PROCEDURE DisplayBoundingBox;
                             PROCEDURE Restart;
                             PROCEDURE AdvanceAnimation ( aTime : INTEGER );
                           END;
  MS3D_Header            = PACKED RECORD
                             Id                  : ARRAY [ 0..9 ] OF CHAR;
                             Version             : INTEGER
                           END;
  MS3D_Vertex            = PACKED RECORD
                             Flags               : BYTE;
                             Position            : ThreeSingles;
                             BoneID              : SHORTINT;
                             refCount            : BYTE
                           END;
  MS3D_Triangle          = PACKED RECORD
                             Flags               : WORD;
                             VertexIndices       : ThreeWords;
                             VertexNormals       : ThreeThreeSingles;
                             s, t                : ThreeSingles;
                             SmoothingGroup,
                             GroupIndex          : BYTE
                           END;
  MS3D_Group             = PACKED RECORD
                             Flags               : BYTE;
                             Name                : ARRAY [ 0..31 ] OF CHAR;
                             nTriangles          : WORD;
                             TriangleIndices     : ARRAY OF WORD;
                             MaterialIndex       : BYTE
                           END;
  MS3D_Material          = PACKED RECORD
                             Name                : ARRAY [ 0..31 ] OF CHAR;
                             Ambient,
                             Diffuse,
                             Specular,
                             Emissive            : FourSingles;
                             Shininess,
                             Transparency        : SINGLE;
                             Mode                : BYTE; //unused!
                             Texture,
                             Alphamap            : ARRAY [ 0..127 ] OF CHAR
                           END;
  MS3D_Joint             = PACKED RECORD
                             Flags               : BYTE;
                             Name,
                             ParentName          : ARRAY [ 0..31 ] OF CHAR;
                             Rotation,
                             Translation         : ThreeSingles;
                             nRotKeyframes,
                             nTransKeyframes     : WORD
                           END;
  MS3D_Keyframe          = PACKED RECORD
                             Time                : SINGLE;
                             Parameter           : ThreeSingles
                           END;
  TJointName             = PACKED RECORD
                             JointIndex          : WORD;
                             Name                : STRING
                           END;
VAR
  Path : STRING;

FUNCTION MS3DLoadStaticModel   ( aPath : STRING ) : TStaticModelGL;
FUNCTION MS3DLoadAnimatedModel ( aPath : STRING ) : TAnimatedModelGL;

IMPLEMENTATION

USES
  Windows, Graphics, Classes, Jpeg;

CONST
  IdentityMatrix: TMatrix = ((1, 0, 0, 0), (0, 1, 0, 0), (0, 0, 1, 0), (0, 0, 0, 1) );
  X = 0;
  Y = 1;
  Z = 2;
  W = 3;

PROCEDURE MatrixRotate ( Matrix : TMatrix; VAR Vector : ThreeSingles );
VAR
  VOrig : ThreeSingles;
BEGIN
  VOrig := Vector;
  Vector [ X ] := VOrig [ X ] * Matrix [ X, X ] +
                  VOrig [ Y ] * Matrix [ Y, X ] +
                  VOrig [ Z ] * Matrix [ Z, X ];
  Vector [ Y ] := VOrig [ X ] * Matrix [ X, Y ] +
                  VOrig [ Y ] * Matrix [ Y, Y ] +
                  VOrig [ Z ] * Matrix [ Z, Y ];
  Vector [ Z ] := VOrig [ X ] * Matrix [ X, Z ] +
                  VOrig [ Y ] * Matrix [ Y, Z ] +
                  VOrig [ Z ] * Matrix [ Z, Z ]
END;

PROCEDURE MatrixInverseRotate ( Matrix : TMatrix; VAR Vector : ThreeSingles );
VAR
  VOrig : ThreeSingles;
BEGIN
  VOrig := Vector;
  vector [ X ] := VOrig [ X ] * Matrix [ X, X ] +
                  VOrig [ Y ] * Matrix [ X, Y ] +
                  VOrig [ Z ] * Matrix [ X, Z ];
  vector [ Y ] := VOrig [ X ] * Matrix [ Y, X ] +
                  VOrig [ Y ] * Matrix [ Y, Y ] +
                  VOrig [ Z ] * Matrix [ Y, Z ];
  vector [ Z ] := VOrig [ X ] * Matrix [ Z, X ] +
                  VOrig [ Y ] * Matrix [ Z, Y ] +
                  VOrig [ Z ] * Matrix [ Z, Z ]
END;


PROCEDURE MatrixSetRotationRad ( VAR Matrix : TMatrix; Param : ThreeSingles );
VAR
  cr, cp, cy, sr, sp, sy : SINGLE;
BEGIN
  cr := cos ( Param [ X ] );
  sr := sin ( Param [ X ] );
  cp := cos ( Param [ Y ] );
  sp := sin ( Param [ Y ] );
  cy := cos ( Param [ Z ] );
  sy := sin ( Param [ Z ] );
  Matrix [ X, X ] := cp * cy;
  Matrix [ X, Y ] := cp * sy;
  Matrix [ X, Z ] := - sp;
  Matrix [ Y, X ] := sr * sp * cy - cr * sy;
  Matrix [ Y, Y ] := sr * sp * sy + cr * cy;
  Matrix [ Y, Z ] := sr * cp;
  Matrix [ Z, X ] := cr * sp * cy + sr * sy;
  Matrix [ Z, Y ] := cr * sp * sy - sr * cy;
  Matrix [ Z, Z ] := cr * cp
END;

PROCEDURE MatrixSetTranslation ( VAR Matrix : TMatrix; Translation : ThreeSingles );
BEGIN
  Matrix [ W, X ] := Translation [ X ];
  Matrix [ W, Y ] := Translation [ Y ];
  Matrix [ W, Z ] := Translation [ Z ]
END;

PROCEDURE MatrixInverseTranslate ( Matrix : TMatrix; VAR Vector : ThreeSingles );
BEGIN
  Vector [ X ] := Vector [ X ] - Matrix [ W, X ];
  Vector [ Y ] := Vector [ Y ] - Matrix [ W, Y ];
  Vector [ Z ] := Vector [ Z ] - Matrix [ W, Z ]
END;


FUNCTION VectorTransform ( Vector: ThreeSingles; Matrix : TMatrix ) : ThreeSingles;
BEGIN
  Result [ X ] := Vector [ X ] * Matrix [ X, X ] +
                  Vector [ Y ] * Matrix [ Y, X ] +
                  Vector [ Z ] * Matrix [ Z, X ] + Matrix [ W, X ];
  Result [ Y ] := Vector [ X ] * Matrix [ X, Y ] +
                  Vector [ Y ] * Matrix [ Y, Y ] +
                  Vector [ Z ] * Matrix [ Z, Y ] + Matrix [ W, Y ];
  Result [ Z ] := Vector [ X ] * Matrix [ X, Z ] +
                  Vector [ Y ] * Matrix [ Y, Z ] +
                  Vector [ Z ] * Matrix [ Z, Z ] + Matrix [ W, Z ]
END;

FUNCTION MatrixMultiply ( Matrix1, Matrix2 : TMatrix ) : TMatrix; REGISTER;
VAR
  i, j : INTEGER;
BEGIN
  FOR i := 0 TO 3 DO
    FOR j := 0 TO 3 DO
      Result[I, J] := Matrix1 [ I, X ] * Matrix2 [ X, J ] +
                      Matrix1 [ I, Y ] * Matrix2 [ Y, J ] +
                      Matrix1 [ I, Z ] * Matrix2 [ Z, J ] +
                      Matrix1 [ I, W ] * Matrix2 [ W, J ]
END;

FUNCTION CreateTexture ( Width, Height, Format : WORD; pData : POINTER ) : INTEGER;
VAR
  Texture : GLuint;
BEGIN
  glGenTextures   ( 1, @Texture );
  glBindTexture   ( GL_TEXTURE_2D, Texture );
  glTexEnvi       ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,   GL_MODULATE );
  glTexParameteri ( GL_TEXTURE_2D,  GL_TEXTURE_MAG_FILTER, GL_LINEAR   );
  glTexParameteri ( GL_TEXTURE_2D,  GL_TEXTURE_MIN_FILTER, GL_LINEAR   );
  IF Format = GL_RGBA
    THEN
      gluBuild2DMipmaps ( GL_TEXTURE_2D, GL_RGBA, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, pData )
    ELSE
      gluBuild2DMipmaps ( GL_TEXTURE_2D, 3, Width, Height, GL_RGB, GL_UNSIGNED_BYTE, pData );
  Result := Texture
END;

FUNCTION LoadJPGTexture ( Filename: STRING; VAR Texture: GLuint; LoadFromResource : BOOLEAN ): BOOLEAN;
VAR
  Data      : ARRAY OF LONGWORD;
  w, Width,
  h, Height : INTEGER;
  BMP       : TBitmap;
  JPG       : TJPEGImage;
  c         : LONGWORD;
  Line      : ^LONGWORD;
  ResStream : TResourceStream;
BEGIN
  Result := FALSE;
  JPG := TJPEGImage.Create;
  IF LoadFromResource
    THEN
      BEGIN
        TRY
          ResStream := TResourceStream.Create ( hInstance, PChar( Copy ( Filename, 1, Pos ('.', Filename ) - 1 ) ), 'JPEG' );
          JPG.LoadFromStream ( ResStream );
          ResStream.Free;
          EXCEPT ON EResNotFound DO
            BEGIN
              MessageBox ( 0, PChar ( 'File not found in resource - ' + Filename ), PChar ( 'JPG Texture' ), MB_OK );
              Exit
            END
          ELSE
            BEGIN
              MessageBox ( 0, PChar('Couldn''t load JPG Resource - "'+ Filename +'"' ), PChar( 'BMP Unit' ), MB_OK );
              Exit
            END
        END
      END
    ELSE
      BEGIN
        TRY
          JPG.LoadFromFile ( Filename );
          EXCEPT
            MessageBox ( 0, PChar ( 'Couldn''t load JPG - "' + Filename + '"' ), PChar ( 'BMP Unit' ), MB_OK );
            Exit
          END
      END;
  BMP := TBitmap.Create;
  BMP.Pixelformat := pf32bit;
  BMP.Width       := JPG.Width;
  BMP.Height      := JPG.Height;
  BMP.Canvas.Draw ( 0, 0, JPG );
  Width  := BMP.Width;
  Height := BMP.Height;
  SetLength ( Data, Width * Height );
  FOR h := 0 TO Height - 1 DO
    BEGIN
      Line :=BMP.Scanline [ Height - h - 1 ];
      FOR w := 0 TO Width - 1 DO
        BEGIN
          c := Line^ AND $FFFFFF;
          Data [ w + ( h * Width ) ] :=( ( ( c AND $FF ) SHL 16 ) + ( c SHR 16 ) + ( c AND $FF00 ) ) OR $FF000000;
          Inc ( Line )
        END
    END;
  BMP.Free;
  JPG.Free;
  Texture := CreateTexture ( Width, Height, GL_RGBA, Addr ( Data [ 0 ] ) );
  Result := TRUE
END;

FUNCTION TStaticModelGL.Render : TModelGLRenderResult;
VAR
  i, j, k : INTEGER;
BEGIN
  FOR i := 0 TO numGroups - 1 DO
    WITH Groups [ i ] DO
      BEGIN
        IF MaterialIndex >= numMaterials
          THEN
            BEGIN
              Result := MATERIAL_NOT_FOUND;
              Exit
            END;
        glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT,  @Materials [ MaterialIndex ].Ambient );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE,  @Materials [ MaterialIndex ].Diffuse );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, @Materials [ MaterialIndex ].Specular );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_SHININESS,@Materials [ MaterialIndex ].Shininess );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, @Materials [ MaterialIndex ].Emissive );
        glBindTexture ( GL_TEXTURE_2D, Materials [ MaterialIndex ].Texture );
        glBegin ( GL_TRIANGLES );
        FOR j := 0 TO nTriangles - 1 DO
          FOR k := 0 TO 2 DO
            WITH Triangles [ TriangleIndices [ j ] ] DO
              BEGIN
                glNormal3fv ( @VertexNormals [ k ] );
                glTexCoord2f ( s [ k ], 1 - t [ k ] );
                glVertex3fv ( @Vertices [ VertexIndices [ k ] ].Position )
              END;
        glEnd
      END;
  Result := RENDERED_OKAY
END;

PROCEDURE TAnimatedModelGL.Restart;
VAR
  i : INTEGER;
BEGIN
  IF numJoints = 0 THEN Exit;
  FOR i := 0 TO numJoints - 1 DO
    BEGIN
      Joints [ i ].CurRotKeyframe := 0;
      Joints [ i ].CurTransKeyframe := 0;
      Joints [ i ].FinalMatrix := Joints [ i ].AbsoluteMatrix
    END;
  Time := MinTime;
END;

PROCEDURE TAnimatedModelGL.DisplayBoundingBox;
VAR
  i      : INTEGER;
  vector : ThreeSingles;
BEGIN
  Vector [ X ] := 0;
  Vector [ Y ] := 0;
  Vector [ Z ] := 0;
  Vector := VectorTransform ( Vector, Joints [ 3 ].FinalMatrix );               // center of Joints
  { calculate and draw bounding box }
  BoundingBoxMin [ X ] := Vector [ X ];
  BoundingBoxMin [ Y ] := Vector [ Y ];
  BoundingBoxMin [ Z ] := Vector [ Z ];
  BoundingBoxMax [ X ] := Vector [ X ];
  BoundingBoxMax [ Y ] := Vector [ Y ];
  BoundingBoxMax [ Z ] := Vector [ Z ];
  FOR i := 1 TO numJoints -1 DO
    BEGIN
      Vector [ X ] := 0;
      Vector [ Y ] := 0;
      Vector [ Z ] := 0;
      Vector := VectorTransform ( Vector, Joints [ i ].FinalMatrix );
      IF Vector [ X ] < BoundingBoxMin [ X ] THEN BoundingBoxMin [ X ] := Vector [ X ];
      IF Vector [ Y ] < BoundingBoxMin [ Y ] THEN BoundingBoxMin [ Y ] := Vector [ Y ];
      IF Vector [ Z ] < BoundingBoxMin [ Z ] THEN BoundingBoxMin [ Z ] := Vector [ Z ];
      IF Vector [ X ] > BoundingBoxMax [ X ] THEN BoundingBoxMax [ X ] := Vector [ X ];
      IF Vector [ Y ] > BoundingBoxMax [ Y ] THEN BoundingBoxMax [ Y ] := Vector [ Y ];
      IF Vector [ Z ] > BoundingBoxMax [ Z ] THEN BoundingBoxMax [ Z ] := Vector [ Z ];
    END;
  BoundingBoxMax [ Y ] := BoundingBoxMax [ Y ] + 0.7;                           // { correct height a bit }
  glColor3f ( 0.3, 0.3, 0.3 );
  glBegin( GL_LINE_LOOP );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMax [ Y ], BoundingBoxMin [ Z ] );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMax [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMax [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMax [ Y ], BoundingBoxMin [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMin [ Y ], BoundingBoxMin [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMin [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMin [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMin [ Y ], BoundingBoxMin [ Z ] );
  glEnd;
  glBegin ( GL_LINES );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMin [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMax [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMin [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMax [ Y ], BoundingBoxMax [ Z ] );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMin [ Y ], BoundingBoxMin [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMin [ Y ], BoundingBoxMin [ Z ] );
    glVertex3f ( BoundingBoxMin [ X ], BoundingBoxMax [ Y ], BoundingBoxMin [ Z ] );
    glVertex3f ( BoundingBoxMax [ X ], BoundingBoxMax [ Y ], BoundingBoxMin [ Z ] );
  glEnd
END;

PROCEDURE TAnimatedModelGL.DisplayBones;
VAR
  i            : INTEGER;
  Vector,
  ParentVector : ThreeSingles;
BEGIN
  IF numJoints = 0 THEN Exit;
  FOR i := 1 TO numJoints - 1 DO
    BEGIN
      Vector [ X ] := 0;
      Vector [ Y ] := 0;
      Vector [ Z ] := 0;
      Vector := VectorTransform ( Vector, Joints [ i ].FinalMatrix );
      glLineWidth ( 1.0 );
      glColor3f ( 1.0, 1.0, 0.0 );
      glBegin ( GL_LINES );
      IF ( Joints [ i ].Parent <> 0 )
        THEN
          BEGIN
            ParentVector [ X ] := 0;
            ParentVector [ Y ] := 0;
            ParentVector [ Z ] := 0;
            Parentvector := VectorTransform ( ParentVector, Joints [ Joints [ i ].Parent ].FinalMatrix );
            glVertex3f ( ParentVector [ X ], ParentVector [ Y ], ParentVector [ Z ] );
            glVertex3f ( Vector [ X ], Vector [ Y ], Vector [ Z ] )
          END;
      glEnd;
      glPointSize ( 2.0 );
      glColor3f ( 0.0, 1.0, 1.0 );
      glBegin ( GL_POINTS );
        glVertex3f ( Vector [ X ], Vector [ Y ], Vector [ Z ] );
      glEnd
    END
END;

FUNCTION TAnimatedModelGL.Render : TModelGLRenderResult;
VAR
  i, j, k, Index : INTEGER;
  BoneID         : INTEGER;
  Vector         : ThreeSingles;
BEGIN
  FOR i := 0 TO numGroups - 1 DO
    WITH Groups [ i ] DO
      BEGIN
        IF MaterialIndex >= numMaterials
          THEN
            BEGIN
              Result := MATERIAL_NOT_FOUND;
              Exit
            END;
        glMaterialfv ( GL_FRONT_AND_BACK, GL_AMBIENT,  @Materials [ MaterialIndex ].Ambient );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_DIFFUSE,  @Materials [ MaterialIndex ].Diffuse );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_SPECULAR, @Materials [ MaterialIndex ].Specular );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_SHININESS,@Materials [ MaterialIndex ].Shininess );
        glMaterialfv ( GL_FRONT_AND_BACK, GL_EMISSION, @Materials [ MaterialIndex ].Emissive );
        glBindTexture ( GL_TEXTURE_2D, Materials [ MaterialIndex ].Texture );
        IF numJoints = 0
          THEN
            BEGIN                                                               // render without animation
              glBegin ( GL_TRIANGLES );
              FOR j := 0 TO nTriangles - 1 DO
                FOR k := 0 TO 2 DO
                  WITH Triangles [ TriangleIndices [ j ] ] DO
                    BEGIN
                      glNormal3fv ( @VertexNormals [ k ] );
                      glTexCoord2f ( s [ k ], 1 - t [ k ] );
                      glVertex3fv ( @Vertices [ VertexIndices [ k ] ].Position )
                    END;
              glEnd
            END
          ELSE
            BEGIN                                                               // render with animation
              glBegin ( GL_TRIANGLES );
              FOR j := 0 TO nTriangles - 1 DO
                FOR k := 0 TO 2 DO
                  WITH Triangles [ TriangleIndices [ j ] ] DO
                    BEGIN
                      Index := VertexIndices [ k ];
                      BoneID := Vertices [ Index ].BoneID;
                      IF BoneID = -1
                        THEN
                          BEGIN
                            glNormal3fv ( @VertexNormals [ k ] );
                            glTexCoord2f ( s [ k ], 1 - t [ k ] );
                            glVertex3fv ( @Vertices [ Index ].Position )
                          END
                        ELSE
                          BEGIN
                            glTexCoord2f ( s [ k ], 1 - t [ k ] );
                            Vector := VertexNormals [ k ];
                            MatrixRotate ( Joints [ BoneID ].FinalMatrix, Vector );
                            glNormal3fv ( @Vector );
                            Vector := ThreeSingles ( VectorTransform ( ThreeSingles ( Vertices [ Index ].Position ),
                                                     Joints [ BoneID ].FinalMatrix ) );
                            glVertex3fv ( @Vector )
                          END
                    END;
              glEnd
            END
      END;
  Result := RENDERED_OKAY;
  glDisable ( GL_LIGHTING );
  IF ShowBoundingBox THEN DisplayBoundingBox;
  glClear   ( GL_DEPTH_BUFFER_BIT );
  IF ShowBones THEN DisplayBones;
  glEnable ( GL_LIGHTING );
END;

PROCEDURE TAnimatedModelGL.AdvanceAnimation ( aTime : INTEGER );
VAR
  i, Frame                   : INTEGER;
  TimeDelta, InterpValue     : SINGLE;
  TransVec, RotVec           : ThreeSingles;
  TransMatrix, RelativeFinal : TMatrix;
  CurFrame, PrevFrame        : TModelKeyframe;
BEGIN
  IF numJoints = 0 THEN exit;
  TransMatrix := IdentityMatrix;
  RelativeFinal := IdentityMatrix;
  Inc( Time, aTime );
  IF Time > MaxTime
    THEN
      BEGIN
        MinTime := 40 * 206;
        MaxTime := 40 * 250;                                                    // idle position
        Restart
      END;
  FOR i := 0 TO numJoints - 1 DO
    BEGIN
      IF ( Joints [ i ].nRotationKeyframes = 0 ) AND ( Joints [ i ].nTranslationKeyframes = 0 )
        THEN
          BEGIN
            Joints [ i ].FinalMatrix := Joints [ i ].AbsoluteMatrix;
            Exit;
          END;
      Frame := Joints [ i ].CurTransKeyframe;
      WHILE ( Frame < Joints [ i ].nTranslationKeyframes ) AND
            ( Joints [ i ].TranslationKeyframes [ Frame ].Time < Time ) DO
        Inc ( Frame );
      Joints [ i ].curTransKeyframe := Frame;
      IF Frame = 0
        THEN
          transVec := Joints [ i ].TranslationKeyframes [ Frame ].Parameter
        ELSE
          IF Frame = Joints [ i ].nTranslationKeyframes
            THEN
              TransVec := Joints [ i ].TranslationKeyframes [ Frame - 1 ].Parameter
            ELSE
              BEGIN
                CurFrame    := Joints [ i ].TranslationKeyframes [ Frame ];
                PrevFrame   := Joints [ i ].TranslationKeyframes [ Frame - 1 ];
                TimeDelta   := CurFrame.Time - PrevFrame.Time;
                InterpValue := ( Time - PrevFrame.Time ) / TimeDelta;
                TransVec [ 0 ] := PrevFrame.Parameter [ 0 ] +
                                 ( CurFrame.Parameter [ 0 ] - prevFrame.Parameter [ 0 ] ) * InterpValue;
                TransVec [ 1 ] := PrevFrame.Parameter [ 1 ] +
                                 ( CurFrame.Parameter [ 1 ] - prevFrame.Parameter [ 1 ] ) * InterpValue;
                TransVec [ 2 ] := PrevFrame.Parameter [ 2 ] +
                                 ( CurFrame.Parameter [ 2 ] - prevFrame.Parameter [ 2 ] ) * InterpValue
              END;
      Frame := Joints [ i ].CurRotKeyframe;
      WHILE ( Frame < Joints [ i ].nRotationKeyframes ) AND
            ( Joints [ i ].RotationKeyframes [ Frame ].Time < Time ) DO
        Inc ( frame );
      Joints [ i ].CurRotKeyframe := frame;
      IF Frame = 0
        THEN
          MatrixSetRotationRad ( TransMatrix, Joints [ i ].RotationKeyframes [ Frame ].Parameter )
        ELSE
          IF Frame = Joints [ i ].nRotationKeyframes
            THEN
              MatrixSetRotationRad ( TransMatrix, Joints [ i ].RotationKeyframes [ Frame - 1 ].Parameter )
            ELSE
              BEGIN
                CurFrame    := Joints [ i ].RotationKeyframes [ Frame ];
                PrevFrame   := Joints [ i ].RotationKeyframes [ Frame - 1 ];
                TimeDelta   := CurFrame.Time - PrevFrame.Time;
                InterpValue := ( Time - PrevFrame.Time ) / TimeDelta;
                RotVec [ 0 ] := PrevFrame.Parameter [ 0 ] +
                               ( CurFrame.Parameter [ 0 ] - prevFrame.Parameter [ 0 ] ) * InterpValue;
                RotVec [ 1 ] := PrevFrame.Parameter [ 1 ] +
                               ( CurFrame.Parameter [ 1 ] - prevFrame.Parameter [ 1 ] ) * InterpValue;
                RotVec [ 2 ] := PrevFrame.Parameter [ 2 ] +
                               ( CurFrame.Parameter [ 2 ] - prevFrame.Parameter [ 2 ] ) * InterpValue;
                MatrixSetRotationRad ( TransMatrix, RotVec )
              END;
      MatrixSetTranslation ( TransMatrix, TransVec );
      RelativeFinal := Joints [ i ].RelativeMatrix;
      RelativeFinal := MatrixMultiply ( TransMatrix, RelativeFinal );
      IF Joints [ i ].Parent = -1
        THEN
          Joints [ i ].FinalMatrix := RelativeFinal
        ELSE
          BEGIN
            Joints [ i ].FinalMatrix := Joints [ Joints [ i ].Parent ].FinalMatrix;
            Joints [ i ].FinalMatrix := MatrixMultiply ( RelativeFinal, Joints [ i ].FinalMatrix )
          END
    END;
END;

PROCEDURE LoadHeader ( aModel : TStaticModelGL; aStream : TStream );            // not used
VAR
  ms3dHeader : MS3D_header;
BEGIN
  WITH aModel DO aStream.Read ( ms3dheader, SizeOf ( ms3dheader ) )
END;

PROCEDURE LoadVertices ( aModel : TStaticModelGL; aStream : TStream );
VAR
  i          : INTEGER;
  ms3dVertex : MS3D_vertex;
BEGIN
  WITH aModel DO
    BEGIN
      aStream.Read ( numVertices, SizeOf ( NumVertices ) );
      SetLength ( Vertices, numVertices );
      FOR i := 0 TO numVertices - 1 DO
        BEGIN
          aStream.Read ( ms3dvertex, SizeOf ( ms3dvertex ) );
          Vertices [ i ].Position := ms3dvertex.Position;
          Vertices [ i ].BoneID   := ms3dvertex.BoneID
        END
    END
END;

PROCEDURE LoadTriangles ( aModel : TStaticModelGL; aStream : TStream );
VAR
  i            : INTEGER;
  ms3dTriangle : MS3D_Triangle;
BEGIN
  WITH aModel DO
    BEGIN
      aStream.Read ( numTriangles, SizeOf ( NumTriangles ) );
      SetLength ( Triangles, numTriangles );
      FOR i := 0 TO NumTriangles - 1 DO
        BEGIN
          aStream.Read( ms3dTriangle, SizeOf ( ms3dTriangle ) );
          Triangles [ i ].VertexIndices    := ms3dTriangle.VertexIndices;
          Triangles [ i ].s                := ms3dTriangle.s;
          Triangles [ i ].t                := ms3dTriangle.t;
          Triangles [ i ].VertexNormals    := ms3dTriangle.VertexNormals
        END
    END
END;

PROCEDURE LoadGroups ( aModel : TStaticModelGL; aStream : TStream );
var
  i, j     : INTEGER;
  ms3dGroup : MS3D_Group;
BEGIN
  WITH aModel DO
    BEGIN
      aStream.Read ( numGroups, SizeOf ( NumGroups ) );
      SetLength ( Groups, numGroups );
      FOR i := 0 TO numGroups - 1 DO
        WITH Groups [ i ] DO
          BEGIN
            aStream.Read ( ms3dGroup.Flags, SizeOf ( ms3dGroup.Flags ) );
            aStream.Read ( ms3dGroup.Name,  SizeOf ( ms3dGroup.Name ) );
            aStream.Read ( nTriangles,      SizeOf ( nTriangles ) );
            SetLength ( TriangleIndices, nTriangles );
            FOR j := 0 TO nTriangles - 1 DO
              aStream.Read ( TriangleIndices [ j ], SizeOf ( TriangleIndices [ j ] ) );
            aStream.Read( MaterialIndex, SizeOf ( MaterialIndex ) )
          END
    END
END;

FUNCTION GetTexturePath ( aModelPath : STRING; aRelTexPath : STRING ) : STRING;
VAR
  i : INTEGER;
BEGIN
  WHILE ( Copy ( aRelTexPath, 0, 3 ) = '..\') AND ( aModelPath <> '' ) DO
    BEGIN
      aRelTexPath := Copy ( aRelTexPath, 4, Length ( aRelTexPath ) );
      i := 0;
      WHILE i <= Length ( aModelPath ) DO
        BEGIN
          IF aModelPath [ i ] = '\' THEN Break;
          Inc ( i )
        END;
      aModelPath := Copy ( aModelPath, i, Length ( aModelPath ) )
    END;
  IF aModelPath = ''
    THEN
      Result := aRelTexPath
    ELSE
      Result := aModelPath + '\' + Copy ( aRelTexPath, 3, Length ( aRelTexPath ) )
END;

PROCEDURE LoadMaterials ( aModel : TStaticModelGL; aStream : TStream );
VAR
  i            : INTEGER;
  TexPath      : STRING;
  ms3dMaterial : MS3D_Material;
BEGIN
  WITH aModel DO
    BEGIN
      aStream.Read ( numMaterials, SizeOf ( NumMaterials ) );
      SetLength ( Materials, numMaterials );
      FOR i := 0 TO NumMaterials - 1 DO
        BEGIN
          aStream.Read ( ms3dmaterial, SizeOf ( ms3dmaterial ) );
          Materials [ i ].Ambient      := ms3dMaterial.Ambient;
          Materials [ i ].Diffuse      := ms3dMaterial.Diffuse;
          Materials [ i ].Specular     := ms3dMaterial.Specular;
          Materials [ i ].Emissive     := ms3dMaterial.Emissive;
          Materials [ i ].Shininess    := ms3dMaterial.Shininess;
          Materials [ i ].Transparency := ms3dMaterial.Transparency;
          IF ms3dMaterial.Texture <> ''
            THEN
              BEGIN
                TexPath := GetTexturePath ( Path, ms3dmaterial.Texture );
                LoadJPGTexture ( TexPath, Materials [ i ].Texture, FALSE )
              END
        END
    END
END;

PROCEDURE SetUpJoints ( aModel : TAnimatedModelGL );
VAR
  i, j  : INTEGER;
  boneID : INTEGER;
  Matrix : TMatrix;
BEGIN
  IF aModel.numJoints = 0 THEN Exit;
  FOR i := 0 TO aModel.numJoints - 1 DO
    WITH aModel.Joints [ i ] DO
      BEGIN
        RelativeMatrix := IdentityMatrix;
        MatrixSetRotationRad ( RelativeMatrix, LocalRotation );
        MatrixSetTranslation ( RelativeMatrix, LocalTranslation );
        IF Parent > -1
          THEN
            BEGIN
              AbsoluteMatrix := aModel.Joints [ Parent ].AbsoluteMatrix;
              AbsoluteMatrix := MatrixMultiply ( RelativeMatrix, AbsoluteMatrix )
            END
          ELSE
            AbsoluteMatrix := RelativeMatrix
      END;
  FOR i := 0 TO aModel.numVertices - 1 DO
    WITH aModel.Vertices [ i ] DO
      IF BoneID > -1
        THEN
          BEGIN
            Matrix :=  aModel.Joints [ BoneID ].AbsoluteMatrix;
            MatrixInverseTranslate ( Matrix, Position );
            MatrixInverseRotate ( Matrix, Position )
          END;
  FOR i := 0 TO aModel.numTriangles - 1 DO
    FOR j := 0 TO 2 DO
      WITH aModel DO
        BEGIN
          BoneID := Vertices [ Triangles [ i ].VertexIndices [ j ] ].BoneID;
          IF BoneID > -1
            THEN
              MatrixInverseRotate ( Joints [ BoneID ].AbsoluteMatrix, Triangles [ i ].Vertexnormals [ j ] )
        END
END;

PROCEDURE LoadAnimation ( aModel : TAnimatedModelGL; aStream : TStream );
VAR
  i, j, k,
  ParentIndex,
  TempStreamPos  : INTEGER;
  TotalFrames    : LONGINT;
  AnimFPS        : SINGLE;
  s              : STRING;
  ms3dJoint      : MS3D_Joint;
  ms3dKeyframe   : MS3D_Keyframe;
  JointNameList  : ARRAY OF TJointName;
BEGIN
  WITH aModel DO
    BEGIN
      aStream.Read ( AnimFPS, SizeOf ( AnimFPS ) );
      aStream.Position := aStream.Position + SizeOf ( Single );                 //Skip CurrentTime
      aStream.Read ( TotalFrames, SizeOf ( TotalFrames ) );
      MaxTime := round ( TotalFrames * 1000 / AnimFPS );
      aStream.Read ( numJoints, SizeOf ( NumJoints ) );
      SetLength ( Joints,numJoints );
      SetLength ( JointNameList, numJoints );
      TempStreamPos := aStream.Position;
      FOR i := 0 TO NumJoints - 1 DO
        BEGIN
          aStream.Read ( ms3dJoint, SizeOf ( ms3dJoint ) );
          aStream.Position := aStream.Position + SizeOf ( ms3dKeyframe ) *
                              ( ms3dJoint.nRotKeyframes + ms3dJoint.nTransKeyframes );
          JointNameList [ i ].JointIndex := i;
          JointNameList [ i ].Name := String ( ms3dJoint.Name )
        END;
      aStream.Position := TempStreamPos;
      ParentIndex := -1;
      FOR i := 0 TO NumJoints-1 DO
        BEGIN
          aStream.Read ( ms3dJoint, SizeOf ( ms3dJoint ) );
          s := String ( ms3dJoint.ParentName );
          IF Length ( s ) > 0
            THEN
              BEGIN
                k := NumJoints;
                REPEAT Dec ( k ) UNTIL ( s = JointNameList [ k ].Name ) OR ( k < 0 );
                IF k >= 0
                  THEN
                    ParentIndex := JointNameList [ k ].JointIndex
                  ELSE
                    MessageBox ( 0, 'Unable to find parent bone in MS3D file', 'Error', MB_OK OR MB_ICONERROR )
              END;
          Joints [ i ].LocalRotation         := ms3dJoint.Rotation;
          Joints [ i ].LocalTranslation      := ms3dJoint.Translation;
          Joints [ i ].parent                := ParentIndex;
          Joints [ i ].nRotationKeyframes    := ms3dJoint.nRotKeyframes;
          Joints [ i ].nTranslationKeyframes := ms3dJoint.nTransKeyframes;
          SetLength ( Joints [ i ].RotationKeyframes,    Joints [ i ].nRotationKeyframes );
          SetLength ( Joints [ i ].TranslationKeyframes, Joints [ i ].nTranslationKeyframes );
          FOR j := 0 TO ms3dJoint.nRotKeyframes - 1 DO
            BEGIN
              aStream.Read ( ms3dKeyframe, SizeOf ( ms3dKeyframe ) );
              Joints [ i ].RotationKeyframes [ j ].JointIndex := i;
              Joints [ i ].RotationKeyframes [ j ].Time       := ms3dKeyframe.Time * 1000;
              Joints [ i ].RotationKeyframes [ j ].Parameter  := ms3dKeyframe.Parameter
            END;
          FOR j := 0 TO ms3dJoint.nTransKeyframes - 1 DO
            BEGIN
              aStream.Read( ms3dKeyframe, SizeOf ( ms3dKeyframe ) );
              Joints [ i ].TranslationKeyframes [ j ].JointIndex := i;
              Joints [ i ].TranslationKeyframes [ j ].Time       := ms3dKeyframe.Time * 1000;
              Joints [ i ].TranslationKeyframes [ j ].Parameter  := ms3dKeyframe.Parameter
            END
        END
    END
END;

FUNCTION ms3dLoadStaticModel ( aPath : STRING ) : TStaticModelGL;
VAR
  i      : INTEGER;
  Stream : TFileStream;
  Model  : TStaticModelGL;
BEGIN
  i := Length ( aPath );
  REPEAT Dec ( i ) UNTIL ( aPath [ i ] = '/' ) OR ( aPath [ i ] = '\' ) OR ( i <= 0 );
  Path := Copy ( aPath, 0, i - 1 );
  Stream := TFileStream.Create ( aPath, 0 );
  Model := TStaticModelGL.Create;
  LoadHeader    ( Model, Stream );
  LoadVertices  ( Model, Stream );
  LoadTriangles ( Model, Stream );
  LoadGroups    ( Model, Stream );
  LoadMaterials ( Model, Stream );
  Stream.Free;
  Result := Model
END;

FUNCTION ms3dLoadAnimatedModel ( aPath : STRING ) : TAnimatedModelGL;
VAR
  i      : INTEGER;
  Stream : TFileStream;
  Model  : TAnimatedModelGL;
BEGIN
  i := Length ( aPath );
  REPEAT Dec ( i ) UNTIL ( aPath [ i ] = '/' ) OR ( aPath [ i ] = '\' ) OR ( i <= 0 );
  Path := Copy ( aPath, 0, i - 1 );
  Stream := TFileStream.Create ( aPath, 0 );
  Model := TAnimatedModelGL.Create;
  LoadHeader    ( Model,Stream );
  LoadVertices  ( Model,Stream );
  LoadTriangles ( Model, Stream );
  LoadGroups    ( Model, Stream );
  LoadMaterials ( Model, Stream );
  LoadAnimation ( Model, Stream );
  setupJoints   ( Model );
  Model.Restart;
  Stream.Free;
  Result := Model
END;

END.
